/* $Id: fclmgr.c,v 4.12 2006/03/17 15:20:31 phils Exp $  */
/* #ident "@(#)vxfs:$RCSfile: fclmgr.c,v $     $Revision: 4.12 $" */

/*
 * $Copyright: Copyright (c) 2009 Symantec Corporation.
 * All rights reserved.
 *
 * THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
 * SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
 * WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
 *
 * The Licensed Software and Documentation are deemed to be commercial
 * computer software as defined in FAR 12.212 and subject to restricted
 * rights as defined in FAR Section 52.227-19 "Commercial Computer
 * Software - Restricted Rights" and DFARS 227.7202, "Rights in
 * Commercial Computer Software or Commercial Computer Software
 * Documentation", as applicable, and any successor regulations. Any use,
 * modification, reproduction release, performance, display or disclosure
 * of the Licensed Software and Documentation by the U.S. Government
 * shall be solely in accordance with the terms of this Agreement.  $
 */

/*
 * FCL Manager:
 * 	Sleeps for a specified time and prints new FCL records on awakening.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>

#ifdef __linux__
#include <stdint.h>
#endif

#include <fcl.h>
#include <vxfsutil.h>

char filename[PATH_MAX];
char arg[PATH_MAX];
char mntpt[PATH_MAX];
char cookiefile[PATH_MAX];
fcl_time_t fcltime;

char *fcltype[] = {
	"No Change",
	"Info_Change",
	"Create",
	"Add_Link",
	"Unlink",
	"Rename",
	"Undelete",
	"Extend",
	"Overwrite",
	"Truncate",
	"Ext_Attr_Chg",
	"Hole_Punch",
	"Header_Rec",
	"Sym_Link",
	"ExtnAttrChg",
	"ReserveChg",
	"ModeChg",
	"OwnerChg",
	"GroupChg",
	"ItimeChg",
	"MtimeChg",
	"AccessInfo",
	"FileOpen",
	"FileClose",
	"FileStats",
	"FclEvntMask",
	NULL
};

static void
sync_fcl_file(
	void)
{
	int error = 0;
	uint64_t loffset;
	if (strlen(mntpt) == 0) {
		printf("Unable to determine mountpoint, will not sync file\n");
		return;
	}

	error = vxfs_fcl_sync(mntpt, &loffset);
	if (error) {
		perror("sync failed");
		return;
	}

	printf("\nSynced the fcl file, offset: %lld\n",
			(unsigned long long)loffset);
}

static char*
getctime(
	time_t *tm)
{
	static char tbuf[100];
	char *ptr = ctime(tm);
	strcpy(tbuf, ptr);
	return tbuf;
}

static void
vx_path_lookup(
	char *mntpt,
	uint64_t inode,
	uint32_t gen,
	char *buf)
{
	char **ino_buf;
	int nentries;
	int i;
	int error = 0;

	error = vxfs_inotopath_gen(mntpt, inode, gen, 0, &ino_buf, &nentries);
	if (error == 0) {
		strcpy(buf, ino_buf[0]);
		for (i = 0; i < nentries; i++) {
			free(ino_buf[i]);
		}
		free(ino_buf);
	} else {

		/*  check for older routine on vxfs 4.0 */
		if (error == ENOTSUP) {
			if ((error = vxfs_inotopath(mntpt, inode, 0, &ino_buf,
							&nentries)) == 0) {
				strcpy(buf, ino_buf[0]);
				for (i = 0; i < nentries; i++) {
					free(ino_buf[i]);
				}
				free(ino_buf);
				return;
			} else if (error == ENOTSUP) {
				printf("\r\nvxfs_inotpath Error, %s",
						strerror(error));
				exit(1);
			}
		}
		strcpy(buf, strerror(error));
	}
}

static void
process_fcl_entry(
	struct fcl_record *fr)
{
	struct fcl_iostats *fiop;
	char buf[PATH_MAX];

	printf("Type of record:%s\n", fcltype[fr->fr_op]);
	printf("Inode number:%llu\t\tInode gen count:%llu\n",
			(unsigned long long)fr->fr_inonum,
			(unsigned long long)fr->fr_inogen);
	printf("Time:%s ", getctime((time_t *)&fr->fr_time.tv_sec));

	/*
	 * fcltime.tv_sec = fr->fr_time.tv_sec;
	 * fcltime.tv_nsec = fr->fr_time.tv_nsec;
	 */

	if(fr->fr_op != VX_FCL_UNLINK) {
		vx_path_lookup(mntpt, fr->fr_inonum, fr->fr_inogen, buf);
		printf("Path to the file:%s\n", buf);
	}

	if (fr->fr_op == VX_FCL_UNLINK || fr->fr_op == VX_FCL_RENAME ||
		fr->fr_op == VX_FCL_LINK) {
		printf("Parent inode number:%llu\t\tParent gen count:%llu\n",
				(unsigned long long)fr->fr_pinonum,
				(unsigned long long)fr->fr_pinogen);
		printf("Filename:%s", fr->fr_filename);

		if (fr->fr_op == VX_FCL_RENAME && fr->fr_newnmvalid) {
			printf("\t\tNew filename:%s\n", fr->fr_newfilename);
		} else {
			printf("\n");
		}
	} else if (fr->fr_op == VX_FCL_FILEOPEN) {
		printf("Command name:%s\n", fr->fr_cmdname);
	} else if (fr->fr_op == VX_FCL_FILESTATS) {
		fiop = fr->fr_stats;

		printf("No. of bytes read:%llu\t\t",
			(unsigned long long)fiop->nbytesread);
		printf("No. of bytes written:%llu\n",
			(unsigned long long)fiop->nbyteswrite);
		printf("Read time:%llu\t\t",
			(unsigned long long)fiop->readtime);
		printf("Write time:%llu\n",
			(unsigned long long)fiop->writetime);
		printf("No. of reads:%llu\t\t",
			(unsigned long long)fiop->nreads);
		printf("No. of writes:%llu\n",
			(unsigned long long)fiop->nwrites);
		printf("Last reset time=%s\n",
			getctime((time_t *) &fiop->lastreset.tv_sec));
	} else if (fr->fr_op == VX_FCL_EVNTMSK_CHG) {
		printf("Old event mask:0x%llx\t\t",
			(unsigned long long)fr->fr_oldmask);
		printf("New event mask:0x%llx\n",
			(unsigned long long)fr->fr_newmask);
	}

	if (fr->fr_acsinfovalid) {
		printf("Effective UID:%u\t\t", fr->fr_acsinfo->fa_euid);
		printf("Effective GID:%u\n", fr->fr_acsinfo->fa_egid);
		printf("Real UID:%u\t\t", fr->fr_acsinfo->fa_ruid);
		printf("Real GID:%u\n", fr->fr_acsinfo->fa_rgid);
		printf("Process ID:%u\n", fr->fr_acsinfo->fa_pid);
		printf("NODE ID:%u\n", fr->fr_acsinfo->fa_nodeid);
	}

	printf("\n");
}

static void
print_usage(
	void)
{
	printf("\nfclmgr <mount_point> <sleep_time> [<cookiefile>]\n");
}

static int
construct_fcl_filename(
	void)
{
	struct stat st;
	int error = 0;
	int i;

	/*
	* if arg is indeed a directory, then construct the fcl filename,
	* else if it is a file, it may be opened directly,
	* else, exit with error
	*/

	/* check if arg exists */
	error = stat(arg, &st);
	if (error) {
		perror("Stat failed");
		exit(error);
	}

	if (S_ISDIR(st.st_mode)) {
		/* construct the path to the fcl file */

		bzero(filename, PATH_MAX);
		strcpy(filename, arg);
		i = strlen(filename);
		if (filename[i-1] != '/')
			filename[i] = '/';
		strcat(filename, VX_FCL_DIR);
		strcat(filename, "/");
		strcat(filename, VX_FCL_NAME);
	} else if (S_ISREG(st.st_mode)) {
		printf("Entered mount point is infact a file .."
			 "please specify a  valid mount point");
		exit(1);
	} else {
		printf("error: %s not a file or directory\n", arg);
		exit(-1);
	}
	printf("fcl filename: %s\n", filename);

	/* verify if the file exists */
	error = stat(filename, &st);
	if (error) {
		perror("Stat of file failed");
		exit(error);
	}

	return error;
}

static int
get_cookie_from_file(
	struct fcl_cookie *fclcook)
{
	int error = 0;
	int nr = 0;
	FILE *fpcook;

	fpcook = fopen(cookiefile, "rb");
	if (fpcook == NULL) {
		return error;
	}

	nr = fread(fclcook, sizeof(*fclcook), 1, fpcook);

	error = fclose(fpcook);
	if (error) {
		return error;
	}
	if (nr != 1) {
		return 1;
	}
	return 0;
}

static int
put_cookie_in_file(
	struct fcl_cookie *fclcook)
{
	int error = 0;
	int nw = 0;
	FILE *fpcook;

	fpcook = fopen(cookiefile, "wb");
	if (fpcook == NULL) {
		return error;
	}

	nw = fwrite(fclcook, sizeof(*fclcook), 1, fpcook);

	error = fclose(fpcook);
	if (error) {
		return error;
	}
	if (nw != 1) {
		return 1;
	}
	return 0;
}

int
main(
	int argc,
	char **argv)
{
	char *str;
	struct stat st;
	int i;
	int stime = 60;
	struct timeval tv;

	void *fh;
	char *bufptr = NULL;
	char buf[8192];
	size_t bufsize = 8192;
	uint64_t eventmask = FCL_ALL_V4_EVENTS;
	uint32_t nentries = 0;
	struct fcl_info fclinfo;

	int error = 0;

	/*
	* option decides the course of action to be taken when
	* when FCL_MISSEDRECORDS is returned.
	* option = 1 : fcl_seek() used to seek to start.
	* option = 2 : fcl_seektime() used.
	* option = 3 : fcl_seek() used to seek to end.
	*/
	int option = 1;

	struct fcl_record *fclrec;
	struct fcl_cookie fclcook;

	if (argc != 3 && argc != 4) {
		print_usage();
		exit(-1);
	}
	str = argv[1];
	bzero(arg, PATH_MAX);
	strcpy(arg, str);

	stime = strtol(argv[2], NULL, 10);
	if (errno) {
		stime = 60;
	}
	printf("Sleep time: %d seconds\n", stime);

	bzero(cookiefile, PATH_MAX);
	if(argc == 4) {
		strcpy(cookiefile, argv[3]);
	} else {
		strcpy(cookiefile, "cookiefile");
	}

	/*
	 * construct the fcl filename
	 */
	if (construct_fcl_filename()) {
		exit(-1);
	}

	/*
	 * get the mount-point
	 */
	bzero(mntpt, PATH_MAX);
	strcpy(mntpt, argv[1]);

	for (;;) {

		/*
		 * open the fcl file
		 */
		error = vxfs_fcl_open(filename, 0, &fh);
		if (error) {
			errno = error;
			perror("Open Failed");
			exit(-1);
		}

		error = vxfs_fcl_getinfo(fh, &fclinfo);
		if (error) {
			perror("vxfs_fcl_getinfo failed");
			exit(-1);
		}
		if (fclinfo.fi_state == VX_FCLS_OFF) {
			printf("fcl has been deactivated,"
				"not printing any records\n");
			exit(-1);
		}

		sync_fcl_file();

		/*
		 * read the cookie from the cookie file.
		 * if no error occurs in reading, seek to cookie.
		 */
		error = get_cookie_from_file(&fclcook);

		if (!error) {
			printf("\nRead cookie.");
			error = vxfs_fcl_seek(fh, &fclcook, FCL_SEEK_COOKIE);

			/*
			 * If FCL_MISSEDRECORDS is returned, either:
			 * (i) activation time stored in cookie is less than
			 * 	the current act. time, or
			 * (ii) offset in cookie is less than the
			 * 	first valid offset in the fcl.
			 * In such a situation, we can either:
			 * (i) seek to the start of the fcl file, or
			 * (ii) seek to the record with time later than
			 *	the time of last record read.
			 * (iii) seek to the end of fcl file
			 */

			if (error == FCL_MISSEDRECORDS) {
				printf("\nMissed some records,");
				printf(" proceeding to read ...\n");
				if (option == 1) {
					error = vxfs_fcl_seek(fh, &fclcook,
							FCL_SEEK_START);
					if (error) {
						errno = error;
						perror("Seek to start failed");
						goto err;
					}
				}
				else if (option == 2) {
					error = vxfs_fcl_seektime(fh, fcltime);
					if (error) {
						errno = error;
						perror("Seek to time failed");
						goto err;
					}
				}
				else if (option == 3) {
					error = vxfs_fcl_seek(fh, &fclcook,
							FCL_SEEK_END);
					if (error) {
						errno = error;
						perror("Seek to end failed");
						goto err;
					}
				}
			} else if (error) {
				errno = error;
				perror("Seek cookie failed");
				goto err;
			}
		}

		for (;;) {
			/*
			 * read the entries from fcl
			 */
			nentries = 0;
			bufptr = buf;
			error = vxfs_fcl_read(fh, bufptr, &bufsize,
					eventmask, &nentries);
			printf("\nNumber of entries read: %d\n", nentries);
			if (error) {
				errno = error;
				perror("Read Failed");
				goto err;
			}
			if (nentries == 0)
				break;

			/*
			 * process the fcl entries
			 */
			while (nentries--) {
				fclrec = (struct fcl_record *)bufptr;
				process_fcl_entry(fclrec);
				bufptr += fclrec->fr_reclen;
			}
		}

		/*
		 * get the time and store it in 'fcltime'
		 */
		error = gettimeofday(&tv,0);
		if (error) {
			errno = error;
			perror("Get time of day failed");
			goto err;
		}
		fcltime.tv_sec = tv.tv_sec;
		fcltime.tv_nsec = tv.tv_usec * 1000;

		error = vxfs_fcl_getcookie(fh, &fclcook);
		if (error) {
			errno = error;
			perror("Get cookie failed");
			goto err;
		}

		/*
		 * write the cookie to the cookie file.
		 */
		error = put_cookie_in_file(&fclcook);
		if (error) {
			printf("\nError occurred in writing to cookie file\n");
			goto err;
		}

err:
		vxfs_fcl_close(fh);
		if (error) {
			break;
		}

		/* sleep for time 'stime' */
		sleep(stime);
	}

	return error;
}

