/* $Id: nattr_cmds.c,v 4.11 2006/03/17 15:20:48 phils Exp $ */
/* #ident "@(#)vxfs:$RCSfile: nattr_cmds.c,v $	$Revision: 4.11 $" */
/*
 * $Copyright: Copyright (c) 2009 Symantec Corporation.
 * All rights reserved.
 *
 * THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
 * SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
 * WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
 *
 * The Licensed Software and Documentation are deemed to be commercial
 * computer software as defined in FAR 12.212 and subject to restricted
 * rights as defined in FAR Section 52.227-19 "Commercial Computer
 * Software - Restricted Rights" and DFARS 227.7202, "Rights in
 * Commercial Computer Software or Commercial Computer Software
 * Documentation", as applicable, and any successor regulations. Any use,
 * modification, reproduction release, performance, display or disclosure
 * of the Licensed Software and Documentation by the U.S. Government
 * shall be solely in accordance with the terms of this Agreement.  $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>

#ifdef __linux__
#include <stdint.h>
#include <string.h>
#include <linux/types.h>
#endif

#include <vxfsutil.h>
#include "common_subs.h"

extern int optind;
extern char *optarg;

static void
usage(
	char	**argv)
{
	printf("Usage:\n");
	printf("%s  -[cdulrtRW] filename  [attr_name  [new_attr_name]]\n",
	       argv[0]);
	printf(" -c : creates attribute for filename\n");
	printf(" -d : lists attribute directory for filename\n");
	printf(" -u : unlinks attribute from filename\n");
	printf(" -l : links attribute for filename\n");
	printf(" -R : reads from attribute for filename\n");
	printf(" -W : writes to attribute for filename\n");
	printf(" -r : renames attribute for filename\n");
	printf(" -s : lists stats for each attribute file\n");
	printf(" -t : Set attribute access/modify time stamp for filename\n");
	exit(1);
}

/*
 * This program is intended to demonstate the usage of the Named Attribute
 * feature of the VxFs 4.0 file system.  The libvxfsutil library provides
 * an application interface for Named Attributes. This program utilizes these
 * API calls to create, delete, and list Named Attributes. Take a look at each
 * branch of the case statement below in the main program for an example of
 * how to use the API calls.
 *
 * About this program
 * This program can be used to create delete and list Named attributes
 * attached to a file. See the "usage" routine above for the command
 * option description.
 * From the command line a single option can be used in the foloowing manner.
 *
 * Example 1
 * Create "attribute1"  on a file "foo"
 *
 * nattr_cmds -c /vxfstest1/foo attribute1
 *
 * Example 2
 * List the attrtribute directory for file foo
 *
 * nattr_cmds -d /vxfstest1/foo
 *
 * Example 3
 * Unlink attribute1 from file "foo"
 *
 * nattr_cmds -u /vxfstest1/foo attribute1
 *
 * Example 4
 * Link attribute2 to attrubte 1 on file foo
 *
 * nattr_cmds -l /vxfstest1/foo attribute1  attribute2
 *
 * Example 5
 * Rename attribute1 to attribute3
 * nattr_cmds -r /vxfstest1/foo attribute1 attribute3
 *
 * Example 6
 * List the attribute directory and file stats for the file foo
 *
 * nattr_cmds -s /vxfstest1/foo
 *
 * Example 7
 * Modify the timestamp for an attribute file, will use today date
 *
 * nattr_cmds -t /vxfstest1/foo attribute1
 *
 *Example 8
 * Write data to the Attribute file
 *
 * nattr_cmds -W /vxfstest1/foo attribute1
 *
 * Example 9
 *
 * Read data from the Attribute file
 *
 * nattr_cmds -R /vxfstest1/foo attribute1
 *
 *
 */

/*
 * Main program for parsing command options and using
 * libvxfsutil API library calls for Named Attribute Feature.
 *
 */
int
main(
	int	argc,
	char	**argv)
{
	int		fd, nfd, err = 0;
	char		*filename;
	char		*attrname;
	char		*attr_new;
	int		mode = 0x777;
	int		attr = 0;
	char		buf[1024];
	char		c;
	char		error_buf[256];
	int		stats = 0;
	int		retval = 0;

	if (argc < 3) {		/* need at least 3 args */
		usage(argv);
	}

	filename = argv[2];

	while ((c = getopt(argc, argv, "cdulrstRW")) != -1) {
		switch (c) {

		case 'c':
			/*
			 * create an attribute
			 */

			if (argc < 4) {	/* need at least 4 args */
				usage(argv);
			}
			attrname = argv[3];

			/*
			 * open regular file, create is necessary.
			 */

			if ((fd = open(filename, O_RDWR | O_CREAT, mode)) < 0) {
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * create named attribute file
			 */

			if ((nfd =
			     vxfs_nattr_open(fd, attrname, O_CREAT,
					     mode)) < 0) {
				sprintf(error_buf,
				    "%s, Error Opening Attribute File %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				close(fd);
				exit(1);
			}

			close(nfd);
			close(fd);

			break;

		case 'u':
			/*
			 * unlink an attribute
			 */

			if (argc < 4) {	/* need at least 4 args */
				usage(argv);
			}
			attrname = argv[3];

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDWR)) < 0) {
				perror("open filename Error:");
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * unlink  attribute
			 */

			if ((err = vxfs_nattr_unlink(fd, attrname)) != 0) {
				sprintf(error_buf,
				     "%s, Error Unlinking Attribute File"
					" %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				close(fd);
				exit(1);
			}

			close(fd);

			break;

		case 'W':
			/*
			 * write data to attribute file
			 */

			if (argc < 4) {	/* need at least 4 args */
				usage(argv);
			}
			attrname = argv[3];

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDWR | O_CREAT, mode)) < 0) {
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * create named attribute file
			 */

			if ((nfd = vxfs_nattr_open(fd, attrname,
						   O_WRONLY | O_CREAT |
						   O_APPEND, mode)) < 0) {
				sprintf(error_buf,
					"%s, Error Opening Attribute File"
						" %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				close(fd);
				exit(1);
			}

			/*
			 * write some data to the  atribute file
			 */

			memset(buf, 0, 1024);
			memset(buf, '?', 1023);
			write(nfd, buf, 1024);

			close(nfd);
			close(fd);
			break;

		case 'R':
			/*
			 * read data from attribute file
			 */

			if (argc < 4) {	/* need at least 4 args */
				usage(argv);
			}
			attrname = argv[3];

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDONLY)) < 0) {
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * open named attribute file
			 */

			if ((nfd =
			     vxfs_nattr_open(fd, attrname, O_RDONLY,
					     mode)) < 0) {

				sprintf(error_buf,
				     "%s, Error Opening Attribute File"
					" %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				close(fd);
				exit(1);
			}

			/*
			 * read some data from the  atribute file
			 */

			read(nfd, buf, 1024);

			printf("Data from file is %s", buf);
			close(nfd);
			close(fd);
			break;

		case 's':
			/*
			 * display stats for each attribute file
			 */

			stats = 1;

		case 'd':
			/*
			 * read the attribute directory
			 */

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDONLY)) < 0) {
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * open the Attributes files directory and get all the
			 * directory entries
			 */

			err = print_nattr_data(fd, filename, mode, stats);

			close(fd);
			break;

		case 'r':
			/*
			 * rename an attribute
			 */

			if (argc < 5) {	/* need at least 5 args */
				usage(argv);
			}
			attrname = argv[3];
			attr_new = argv[4];

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDWR)) < 0) {
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * rename attribute
			 */

			if ((err =
			     vxfs_nattr_rename(fd, attrname, attr_new)) != 0) {
				sprintf(error_buf,
				     "%s, Error renaming Attribute File"
					" %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				exit(1);
			}
			close(fd);
			break;

		case 'l':
			/*
			 * link an attribute file
			 */

			if (argc < 5) {	/* need at least 5 args */
				usage(argv);
			}
			attrname = argv[3];
			attr_new = argv[4];

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDWR)) < 0) {
				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * link attribute
			 */

			if ((err =
			     vxfs_nattr_link(fd, attrname, attr_new)) != 0) {

				sprintf(error_buf,
				    "%s, Error linking Attribute File %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				exit(1);

			}
			close(fd);
			break;

		case 't':
			/*
			 * set atime/mtime  for attribute file
			 */

			if (argc < 4) {	/* need at least 4 args */
				usage(argv);
			}
			attrname = argv[3];

			/*
			 * open file
			 */

			if ((fd = open(filename, O_RDWR)) < 0) {

				sprintf(error_buf, "%s, Error Opening File %s ",
					argv[0], filename);
				perror(error_buf);
				exit(1);
			}

			/*
			 * set attribute atime/mtime to current TOD
			 * NOTE: Using NULL for the 3rd arguement will
			 * cause the timestamp to be the current time.
			 */

			if ((err = vxfs_nattr_utimes(fd, attrname,NULL)) != 0) {

				/*
				 * An alternate way to modify the time stamp,
				 * is to pass a timeval structure as the 3rd
				 *  arguement
				 * struct timeval time;
				 *  gettimeofday(&time, NULL);
				 *
				 * if ( (err =
				 *    vxfs_nattr_utimes(fd, attrname, time)) ) {
				 */

				sprintf(error_buf,
				"%s, Error setting timestamp on Attribute File"
					" %s/./%s ",
					argv[0], filename, attrname);
				perror(error_buf);
				exit(1);

			}
			close(fd);
			break;

		default:
			usage(argv);
			break;
		}
	}
	exit(0);
}
