/* $Id: freeze.c,v 4.2 2006/03/17 15:26:30 phils Exp $ */
/* #ident "@(#)vxfs:$RCSfile: freeze.c,v $	$Revision: 4.2 $" */

/*
 * $Copyright: Copyright (c) 2009 Symantec Corporation.
 * All rights reserved.
 *
 * THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
 * SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
 * WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
 *
 * The Licensed Software and Documentation are deemed to be commercial
 * computer software as defined in FAR 12.212 and subject to restricted
 * rights as defined in FAR Section 52.227-19 "Commercial Computer
 * Software - Restricted Rights" and DFARS 227.7202, "Rights in
 * Commercial Computer Software or Commercial Computer Software
 * Documentation", as applicable, and any successor regulations. Any use,
 * modification, reproduction release, performance, display or disclosure
 * of the Licensed Software and Documentation by the U.S. Government
 * shall be solely in accordance with the terms of this Agreement.  $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/fs/vx_ioctl.h>
#include <mntent.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>

#define MNTAB "/etc/mtab"

/* Global Variables */
int	vxfs_fd;
int	success_flag = 0;
pid_t	parent = 0;

/* Signal Handlers */

/*
 * parent's handlers
 */

void
success(
	int i)
{
	printf("\r\nChild process has successfully Frozen File System\n");
	success_flag = 1;
}

void
failure(
	int i)
{
	printf("Child Process failed to Freeze File System\n");
	exit(1);
}

 /*
  * used by parent and child to thaw file system
  */

void
thaw(
	int i)
{
	if (ioctl(vxfs_fd, VX_THAW, NULL)) {
		perror("ERROR: File system thaw failed");
		close(vxfs_fd);
		exit(1);
	} else {
		printf("\r\n%s process has successfully"
		       " Thaw'ed File system\n",
		       (getpid() == parent) ? "Parent" : "Child");
		close(vxfs_fd);
		exit(0);
	}
}

/* Usage function */
void
usage(
	char *command)
{
	fprintf(stderr, "USAGE: %s [ -t timeout ] filesystem_mount_point\n",
		command);
	exit(1);
}

static void
print_menu(void)
{
	printf("\nMenu choices\n");

	printf("\n t  - Early Thaw - Thaw File system");
	printf("\n q  - Wait for Child to complete Freeze/Thaw File system");

	printf("\n    Enter Choice ==> ");
}

int
main(
	int	argc,
	char	**argv)
{
	long		timeout = 30;
	uid_t		iam = geteuid();
	pid_t		gppid = getpid();
	int		opt, status;
	char		*mntpt;
	struct mntent	*mnttab_entry;
	FILE		*mnttabfile;
	int		i;
	char		choice;
	pid_t		child;
	int		entry_found = 0;

	/*
	 * Parse the args
	 */

	while ((opt = getopt(argc, argv, "t:")) != EOF) {
		switch (opt) {
		case 't':
			timeout = atol(optarg);
			break;

		case '?':
			usage(argv[0]);
			break;
		}
	}
	mntpt = argv[argc - 1];

	/*
	 * Make sure they entered enough arguments, otherwise print usage
	 */

	if (argc < 2) {
		usage(argv[0]);
	}

	/*
	 * Check the timeout
	 */

	if ((timeout < 10) || (timeout > 60)) {
		fprintf(stderr,
			"ERROR: Timeout must be between 10 and 60 seconds"
			" (default 30).\n");
		exit(1);
	}

	/*
	 * Verify that the mount point is an absolute path
	 */

	if (strncmp(mntpt, "/", 1)) {
		fprintf(stderr,
			"ERROR: File system mount point must begin with /.\n");
		exit(1);
	}

	/*
	 * Make sure that we are root
	 */

	if (iam != 0) {
		fprintf(stderr, "ERROR: %s must be run setuid/as root.\n",
			argv[0]);
		exit(1);
	}

	/*
	 * Get the mnttab entry
	 */

	mnttabfile = setmntent(MNTTAB, "r");
	if (!mnttabfile) {
		perror("ERROR: Unable to open the MNTTAB");
		exit(1);
	}

	while ((mnttab_entry = getmntent(mnttabfile)) != NULL) {
		if ((strcmp (mnttab_entry->mnt_dir, mntpt) == 0) && \
		    (strcmp (mnttab_entry->mnt_type, "vxfs") == 0)) {
				entry_found = 1;
		}
	}
	endmntent(mnttabfile);
	if (!entry_found) {
		fprintf(stderr,
			"ERROR: file system %s not found"
			" in MNTTAB or is not VXFS.\n", mntpt);
		exit(1);
	}

	/*
	 * Child does the work, Parent waits for signal to return status
	 */

	/*
	 * if open is done here, then the parent & child can THAW
	 * since child inherits fd
	 */

	vxfs_fd = open(mntpt, O_RDONLY);
	parent = getpid();

	if ((child = fork()) == 0) {

		/*
		 * Child, setup the signal handlers
		 */

		signal(SIGUSR2, thaw);

		/*
		 * Freeze filesystem, signal parent to return status
		 */

		/*
		 * if open is done here, only the child can THAW
		 */

		if (vxfs_fd < 0) {
			perror("ERROR: Failed to open device file");
			exit(1);
		}
		if (ioctl(vxfs_fd, VX_FREEZE, timeout)) {
			perror("ERROR: File system freeze failed");
			close(vxfs_fd);
			kill(gppid, SIGUSR2);
			exit(1);
		} else {
			kill(gppid, SIGUSR1);

			/*
			 * Print our pid, then wait for later thaw signal
			 */

			/*
			 * fclose(stdout);
			 */

			sleep(timeout);
			close(vxfs_fd);
			exit(0);
		}
	} else {

		/*
		 * Parent
		 */

		/*
		 * setup parent keyboard handlers
		 */

		signal(SIGHUP, thaw);
		signal(SIGINT, thaw);
		signal(SIGTERM, thaw);

		/*
		 * Setup to receive success/failure signal from child
		 */

		signal(SIGUSR1, success);
		signal(SIGUSR2, failure);

		/*
		 * wait for chiled to freeze
		 */

		printf("\nWaiting for Child to Freeze File system\n");
		while (!success_flag);

		while (1) {
			print_menu();
			choice = getchar();
			switch (choice) {

			case 't':
				/*
				 * send a signal thaw the file system
				 * before the timeout period
				 */

				kill(child, SIGUSR2);

			case 'q':
				goto out;
				break;

			default:
				break;
			}

			while ((choice = getchar()) != '\n');
		}

	      out:
		wait(&status);
		exit(0);
	}
}
